use std::{
    cmp::Ordering,
    fmt::{self, Display, Formatter},
};

use serde::Serialize;

#[derive(Debug, Serialize)]
pub struct Fraction {
    pub numerator: i64,
    pub denominator: i64,
}

impl Fraction {
    pub fn new_from_float(numerator: f64, denominator: f64) -> Self {
        if denominator == 0.0 {
            panic!("denomitator cannot be zero");
        }

        // 计算转换因子，以消除浮点数的小数部分
        let scale_factor = 10i64
            .pow(Self::decimal_places(numerator).max(Self::decimal_places(denominator)) as u32);

        let int_numerator = (numerator * scale_factor as f64).round() as i64;
        let int_denominator = (denominator * scale_factor as f64).round() as i64;

        let gcd = Self::gcd(int_numerator, int_denominator);
        Fraction {
            numerator: int_numerator / gcd,
            denominator: int_denominator / gcd,
        }
    }

    pub fn new(numerator: i64, denominator: i64) -> Self {
        if denominator == 0 {
            panic!("denominator cannot be zero");
        }

        let gcd = Self::gcd(numerator, denominator);
        Fraction {
            numerator: numerator / gcd,
            denominator: denominator / gcd,
        }
    }

    fn decimal_places(value: f64) -> u32 {
        let mut decimal_places = 0;
        let mut scaled_value = value;
        while scaled_value.fract() != 0.0 && decimal_places < 15 {
            scaled_value *= 10.0;
            decimal_places += 1;
        }
        decimal_places
    }

    fn gcd(a: i64, b: i64) -> i64 {
        let mut a = a.abs();
        let mut b = b.abs();
        while b != 0 {
            let temp = b;
            b = a % b;
            a = temp;
        }
        a
    }

    fn cmp(&self, other: &Fraction) -> Ordering {
        (self.numerator * other.denominator).cmp(&(other.numerator * self.denominator))
    }
}

impl PartialEq for Fraction {
    fn eq(&self, other: &Self) -> bool {
        self.cmp(other) == Ordering::Equal
    }
}

impl PartialOrd for Fraction {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl Display for Fraction {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        write!(f, "{}/{}", self.numerator, self.denominator)
    }
}
