mod algorithms;
mod frequency_shifting;
mod pulse;

use std::collections::VecDeque;

use chrono::{
    naive::serde::{ts_milliseconds, ts_milliseconds_option},
    NaiveDateTime,
};
pub use pulse::Pulse;
use pulse::{generate_pulse_sequence, RawPulse};
use serde::{Deserialize, Serialize};
use sled::IVec;
use uuid::Uuid;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Pattern {
    pub id: Uuid,
    pub name: String,
    #[serde(with = "ts_milliseconds")]
    pub created_at: NaiveDateTime,
    #[serde(with = "ts_milliseconds_option")]
    pub last_modified_at: Option<NaiveDateTime>,
    pub smooth_repeat: bool,
    pub pulses: VecDeque<Pulse>,
}

impl Pattern {
    pub fn add_pulse(&mut self, pulse: Pulse) {
        self.pulses.push_back(pulse);
    }

    pub fn delete_pulse(&mut self, pulse_id: &Uuid) {
        if let Some(index) = self.pulses.iter().position(|p| &p.id == pulse_id) {
            self.pulses.remove(index);
        }
    }

    pub fn move_forward(&mut self, pulse_id: &Uuid) {
        if let Some(index) = self.pulses.iter().position(|p| &p.id == pulse_id) {
            if index > 0 && index <= self.pulses.len() - 1 {
                self.pulses.swap(index, index - 1);
            }
        }
    }

    pub fn move_backward(&mut self, pulse_id: &Uuid) {
        if let Some(index) = self.pulses.iter().position(|p| &p.id == pulse_id) {
            if index < self.pulses.len() - 1 {
                self.pulses.swap(index, index + 1);
            }
        }
    }

    pub fn set_smooth_repeat(&mut self, repeat: bool) {
        self.smooth_repeat = repeat;
    }

    pub fn output_pulses(&self) -> Box<dyn Iterator<Item = RawPulse>> {
        generate_pulse_sequence(self.pulses.clone(), self.smooth_repeat)
    }

    pub fn total_duration(&self) -> u32 {
        let mut duration = 0;
        for (index, pulse) in self.pulses.iter().enumerate() {
            if (index == self.pulses.len() - 1 && self.smooth_repeat)
                || index < self.pulses.len() - 1
            {
                duration += pulse.duration;
            }
        }
        duration
    }
}

impl From<&[u8]> for Pattern {
    fn from(bytes: &[u8]) -> Self {
        bincode::deserialize(&bytes).unwrap()
    }
}

impl From<IVec> for Pattern {
    fn from(bytes: IVec) -> Self {
        bincode::deserialize(&bytes).unwrap()
    }
}

impl Into<IVec> for Pattern {
    fn into(self) -> IVec {
        bincode::serialize(&self).unwrap().into()
    }
}

impl Into<IVec> for &Pattern {
    fn into(self) -> IVec {
        bincode::serialize(self).unwrap().into()
    }
}
